import sys
from time import sleep


def print_all_colours():
    for i in range(0, 16):
        for j in range(0, 16):
            code = str(i * 16 + j)
            body = "\033[38;5;" + code + "m "
            sys.stdout.write("\033[1m" + body + code.ljust(4))
        print("\u001b[0m")


def clr(texts: tuple, colors: tuple, bold=True) -> str:
    """Add color codes to string"""
    string = ""
    if texts.__len__() != colors.__len__():
        raise ValueError("Missiong text or color")
    if bold:
        bold = "\033[1m"
    else:
        bold = ""
    for i, text in enumerate(texts):
        body = "\033[38;5;" + str(colors[i]) + "m"
        end = "\033[0m"
        string += bold + body + f"{text} " + end
    return string


def inverse(text):
    return f"\033[;7m{text}\033[0m"


def color_bg(text, fg="", bg=""):
    if not fg:
        fg = 1
    else:
        fg = "3" + str(fg)
    if not bg:
        bg = 2
    else:
        bg = "4" + str(bg)
    return f"\033[{fg};{bg}m{text}\033[m"


def rgb(text, r, g, b, bold=True):
    if bold:
        bold = "\033[1m"
    else:
        bold = ""
    return f"{bold}\033[38;2;{r};{g};{b}m{text} \033[0m"


class Cursor:
    @staticmethod
    def up(lines: int):
        sys.stdout.write("\033[F" * lines)

    @staticmethod
    def down(lines: int):
        sys.stdout.write("\033[E" * lines)

    @staticmethod
    def column(position: int):
        sys.stdout.write(f"\033[{position}G")

    @staticmethod
    def beginning():
        sys.stdout.write("\r")


if __name__ == "__main__":
    print_all_colours()
    print(inverse("Text"))
    print(color_bg("Color BG", 1, 2))
    print(clr(("text", "bext"), (216, 3)))
    print(rgb("Hello, World", 102, 255, 178, True))

    print("Line 1")
    print("Line 2")
    sleep(1)
    Cursor.up(2)
    Cursor.column(3)
    print("Hello World")
    sleep(1)
    Cursor.down(3)
    print("Zebra", end="")
    sleep(3)
    Cursor.beginning()
    print("Bebra")
    for i in range(20):
        print(f"{i}    ", end="\r")
        sleep(0.5)

    input("Enter a 5 letter word: ")
    Cursor.up(1)
    print("words go here          ")
    # print('\x1b[1A' + 'words go here          ')



import json


def sort_dict_by_keys(data: dict):
    return dict(sorted(data.items(), key=lambda x: x[0]))


def sort_nested_dict_by_values(data: dict):
    return json.loads(json.dumps(data, sort_keys=True))


def sort_dict_by_values(data: dict):
    return dict(sorted(data.items(), key=lambda x: x[1]))



import os
import re


def load_dotenv(path: str = ".env"):
    to_strip = ["\n", "\t", "\r", " ", "'", '"']
    with open(path) as f:
        for line in f:
            if line != "\n" and not line.startswith("#"):
                key = re.findall(r"(.*?) ?=", line)[0]
                key = key.strip("".join(to_strip))
                value = re.findall(r".*?= ?(.*)", line)[0]
                value = value.strip("".join(to_strip))
                if key and value:
                    os.environ.update({key: value})


if __name__ == "__main__":
    load_dotenv(".env")



import sys
from pathlib import Path
from time import sleep

sys.path.append(str(Path(__file__).parent.parent))
from src import ytils


def test_printer():
    from src.ytils import printer

    print("Line 1")
    print("Line 2")
    sleep(1)
    printer.Cursor.up(2)
    printer.Cursor.column(3)
    print("Hello World")
    sleep(1)
    printer.Cursor.down(3)
    print("Zebra", end="")
    sleep(3)
    printer.Cursor.beginning()
    print("Bebra")
    # for i in range(20):
    #     print(f"{i}    ", end="\r")
    #     sleep(0.5)

    input("Enter a 5 letter word: ")
    printer.Cursor.up(1)
    print("words go here          ")
    # print('\x1b[1A' + 'words go here          ')


def test_json():
    import json
    from datetime import datetime

    from src.ytils.json import DateTimeEncoder, decode_datetime

    with open("data.json", "w", encoding="utf-8") as f:
        f.write(json.dumps({"datetime": datetime.now()}, cls=DateTimeEncoder, ensure_ascii=False))
    with open("data.json") as f:
        data = json.loads(f.read(), object_hook=decode_datetime)
        print(data)


if __name__ == "__main__":
    test_json()
    # test_printer()



import json
from datetime import datetime


class DateTimeEncoder(json.JSONEncoder):
    """json.dumps(data, cls=DateTimeEncoder)"""
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)


def decode_datetime(dct):
    """json.loads(f.read(), object_hook=decode_datetime)"""
    for key, value in dct.items():
        try:
            # Attempt to parse the value as a datetime string
            dct[key] = datetime.fromisoformat(value)
        except (ValueError, TypeError):
            # If it's not a valid datetime string, leave it as is
            pass
    return dct



import base64
import io
import re
import sys
from typing import Tuple


def PIL_image_to_base64(img, format="PNG") -> bytes:
    img_buffer = io.BytesIO()
    img.save(img_buffer, format=format)
    return base64.b64encode(img_buffer.getvalue())


def bytes_into_unit(bytes: float, units=["B", "KB", "MB", "GB", "TB"], precision=1) -> str:
    """Converts bytes into human readable size"""
    for unit in units:
        if bytes < 1024.0:
            break
        bytes /= 1024.0
    return f"{bytes:.{precision}f} {unit}"


def seconds_to(sec: int, *, to="hours|minutes") -> Tuple[int, int, int]:
    """
    Returns hours, minutes, seconds
    """
    assert to == "hours" or to == "minutes", "to can be hours or minutes"

    result = []
    min, sec = divmod(sec, 60)
    if "hours" in to:
        hrs, min = divmod(min, 60)
        result.append(hrs)
    else:
        result.append(0)

    result.append(min)
    result.append(sec)

    return tuple(result)


def sanitize_file_name(file_name: str, replace_with="_") -> str:
    """Remove prohibited symbols from file name"""
    prohibited_symbols = r'[\\/:*?"<>|]'  # List of prohibited symbols
    sanitized_file_name = re.sub(prohibited_symbols, replace_with, file_name)

    return sanitized_file_name


def image_to_base64(path: str, compress: int = None, markdown=False) -> str:
    if "Image" not in sys.modules:
        from PIL import Image

    image = Image.open(path)
    width, height = image.size
    print("Size:", f"{width}x{height}")
    if compress:
        new_size = (width // compress, height // compress)
        image = image.resize(new_size)
    buffered = io.BytesIO()
    image.save(buffered, format="PNG")
    b64 = base64.b64encode(buffered.getvalue())
    if markdown:
        return f"![Name](data:image/png;base64,{b64.decode()})"
    else:
        return base64.decode()


if __name__ == "__main__":
    sec = 6432
    h, m, s = seconds_to(sec, "hours")  # (1, 47, 12)
    print(h, m, s)
    h, m, s = seconds_to(sec, "minutes")  # (0, 107, 12)
    print(h, m, s)



from . import converters
from . import dotenv
from . import logger
from . import obfuscator
from . import perfomance
from . import printer
from . import reg_exp
from . import sorters
from . import web
from . import threads
from . import json


__version__ = "0.0.3"



import datetime
import logging
import os
from collections import deque
from pathlib import Path


class Logger(logging.Logger):

    def __init__(self, name=__name__, level=logging.INFO, file: str | bool = False, *args, **kwargs):
        """Change logger configuration here"""
        logging.basicConfig(encoding="utf-8")
        super().__init__(name=name, level=level, *args, **kwargs)
        format = "%(asctime)s %(levelname)+6s %(name)s %(filename)s(%(lineno)s): %(message)s"
        formatter = logging.Formatter(format, "%H:%M:%S")

        # Write logs to std
        streamHandler = logging.StreamHandler()
        streamHandler.setFormatter(formatter)
        self.addHandler(streamHandler)

        # Write logs to file
        if file:
            now = datetime.datetime.now()
            if file == True:
                logs_dir = Path("logs")
                logs_dir.mkdir(exist_ok=True)
                file = logs_dir

            if Path(file).is_dir():
                logs_dir = Path(file)
                file = logs_dir / (f'{name}_{now.strftime("%Y-%m-%d_%H-%M-%S")}.log')

            fileHandler = logging.FileHandler(file, encoding="utf-8")
            fileHandler.setFormatter(formatter)
            self.addHandler(fileHandler)


class _SingletonType(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_SingletonType, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SingletoneLogger(Logger, metaclass=_SingletonType):
    """Returnes the same instance every time"""

    def __init__(self, name=__name__, level=logging.INFO, file: str = False):
        super().__init__(name, level, file)


class LimitedLinesFileHandler(logging.FileHandler):
    """
    Limits max file length by trimming the file to max_lines lines.

    Usage:
    file_handler = LimitedLinesFileHandler("app_lock.log", mode="a", encoding="utf-8")
    """

    def __init__(self, filename, mode="a", encoding=None, delay=False, errors: str = None, max_lines=10_000):
        super().__init__(filename, mode, encoding, delay, errors)
        self.max_lines = max_lines

    def emit(self, record):
        super().emit(record)
        self._trim_file()

    def _trim_file(self):
        try:
            self.stream.flush()
            with open(self.baseFilename, "r", encoding=self.encoding) as f:
                # Use deque to keep only the last max_lines lines
                lines = deque(f, maxlen=self.max_lines)

            # Write the trimmed lines back to the file
            with open(self.baseFilename, "w", encoding=self.encoding) as f:
                f.writelines(lines)
        except Exception:
            self.handleError(None)


if __name__ == "__main__":
    # Set basic config
    logging.basicConfig(level=logging.DEBUG, encoding="utf-8")
    # Turn off specific logger
    logging.getLogger("aiogram").setLevel(logging.CRITICAL)
    # Usual logger
    logger = Logger()
    logger.error("Hello")
    logger = Logger("ytils.logger")
    logger.error("Hello 2")
    # Singletone logger
    logger = SingletoneLogger("singletone")
    logger.error("Hello 3")
    logger = SingletoneLogger("singletone2")
    logger.error("Hello 4")



import logging
import time
from functools import wraps

from .logger import Logger

logger = Logger("ytils." + __name__)


def track_time(log_fun):
    def _track_time(fn):
        @wraps(fn)
        def wrapped_fn(*args, **kwargs):
            try:
                start_time = time.time()
                result = fn(*args, **kwargs)
            finally:
                elapsed_time = time.time() - start_time
                log_fun({"fn_name": fn.__name__, "elapsed_time": elapsed_time})
            return result

        return wrapped_fn

    return _track_time


class Profiler:

    def __init__(self, logger: logging.Logger = logger) -> None:
        self.logger = logger
        # CPU time.process_time()
        self._s = time.perf_counter()  # Start time
        self._m = self._s  # Marker for time since the last trace

    def restart(self):
        self._s = time.perf_counter()
        self._m = self._s

    def trace(self, text: str, restart=False):
        current_time = time.perf_counter()
        elapsed_since_start = current_time - self._s
        elapsed_since_marker = current_time - self._m
        self.logger.debug(f"Profiler:{text.rjust(50)}: {elapsed_since_marker:.5f} total: {elapsed_since_start:.5f}")
        self._m = current_time
        if restart == True:
            self.restart()


@track_time(lambda message: logger.info("{elapsed_time:.3f}s > {fn_name}".format(**message)))
def _test_track_time():
    time.sleep(5)


def _test_profiler():
    profiler = Profiler()
    time.sleep(3)
    profiler.trace("1")
    time.sleep(5)
    profiler.trace("2", restart=True)
    time.sleep(2)
    profiler.trace("3", restart=True)


if __name__ == "__main__":
    _test_track_time()
    _test_profiler()



import base64


def _obfuscate(text: bytes, mask: bytes) -> bytes:
    lmask = len(mask)
    return bytes(c ^ mask[i % lmask] for i, c in enumerate(text))


def obfuscate(text, mask):
    b64_byte_text = base64.b64encode(text.encode())
    return _obfuscate(b64_byte_text, mask.encode()).decode("ISO-8859-1")


def deobfuscate(text, mask):
    b64_byte_text = _obfuscate(text.encode("ISO-8859-1"), mask.encode())
    return base64.b64decode(b64_byte_text).decode()


if __name__ == "__main__":
    mask = "Mask"
    text = "Text123456&свсыхцёэ'іїхсвсячюбс😀"

    text = obfuscate(text, mask)
    print(text)
    print(deobfuscate(text, mask))



import threading


class StoppableThread(threading.Thread):

    def __init__(
        self,
        group=None,
        target: callable = None,
        name: str = None,
        args: iter = (),
        kwargs: dict = None,
        *,
        daemon: bool = None,
    ):
        super().__init__(group, target, name, args, kwargs, daemon=daemon)
        self._stop_event = threading.Event()

    def run(self):
        if self._target is not None:
            self._target(self, *self._args, **self._kwargs)
        else:
            raise Exception("Missing target")

    def stop(self):
        self._stop_event.set()

    def is_stopped(self):
        return self._stop_event.is_set()


class ThreadWithResult(threading.Thread):

    def __init__(
        self,
        group=None,
        target: callable = None,
        name: str = None,
        args: iter = (),
        kwargs: dict = None,
        *,
        daemon: bool = None,
    ) -> None:
        super().__init__(group, target, name, args, kwargs, daemon=daemon)
        self._result = None

    def run(self):
        if self._target is not None:
            self._result = self._target(*self._args, **self._kwargs)
        else:
            raise Exception("Missing target")

    def result(self, start=False, join=False):
        if start:
            self.start()
        if join:
            self.join()
        return self._result


if __name__ == "__main__":
    import time
    t = ThreadWithResult(target=lambda x: x if time.sleep(5) else x * 2, args=(7,))
    t.start()
    t.join()
    result = t.result()
    print(result)



import re
import traceback

import requests
try:
    from fake_useragent import FakeUserAgent
    from linkpreview import Link, LinkGrabber, LinkPreview, link_preview
except ImportError:
    pass

from .logger import Logger
from .reg_exp import RE

logger = Logger("ytils." + __name__)


def get_preview(url: str, content: str = None, parser: str = "html.parser") -> "LinkPreview":
    """
    The same as link_preview but with bigger buff size
    parser can be 'lxml'
    """
    if content is None:
        grabber = LinkGrabber(
            initial_timeout=50,
            maxsize=9048576,
            receive_timeout=50,
            chunk_size=1024,
        )
        content, url = grabber.get_content(url)

    link = Link(url, content)
    return LinkPreview(link, parser=parser)


def get_user_agent() -> str:
    default_user_agent = "Mozilla/5.0 (Linux; arm_64; Android 9; SM-A505FN) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 YaApp_Android/20.113.0 YaSearchBrowser/20.113.0 BroPP/1.0 SA/3 Mobile Safari/537.36"

    if not hasattr(get_user_agent, "fake_user_agent"):
        try:
            fake_user_agent = FakeUserAgent()
            get_user_agent.fake_user_agent = fake_user_agent
        except Exception as e:
            logger.error(traceback.format_exc())
            logger.error(e)

    if hasattr(get_user_agent, "fake_user_agent"):
        try:
            user_agent = get_user_agent.fake_user_agent.random
        except:
            user_agent = default_user_agent
    else:
        user_agent = default_user_agent

    return user_agent


def get_web_page_title(url: str) -> str:
    title = ""
    try:
        for i in range(3):
            user_agent = get_user_agent()
            headers = {"User-Agent": user_agent}
            if "youtube" in url:
                headers = {}

            response = requests.get(url, headers=headers)
            content = response.content
            if isinstance(content, bytes):
                content = content.decode()

            if len(content) != 0:
                search_results = re.search(r"<\W*title\W*(.*)</title", content, re.IGNORECASE)
                if not search_results:
                    continue
                title = search_results.group(1)
                if ">" in title:
                    x = title.split(">")[1:]
                    title = ">".join(x)
                if title:
                    break
    except Exception as e:
        logger.error(traceback.format_exc())
    return title


def extract_links_from_text(text: str) -> list:
    result = re.findall(RE.LINK, text)
    urls = [x[0] for x in result]
    return urls


def save_response_html(response, path="index.html"):
    with open(path, "wb") as f:
        f.write(response.content)

    return response.text


def get_ipinfo() -> dict:
    """Get user latitude and longitude etc"""
    try:
        response = requests.get('https://ipinfo.io')
        data = response.json()
        return data
    except:
        print("Error: Unable to detect your location.")
        return None

if __name__ == "__main__":
    text = "Hello https://play.google.com/store/apps/details?id=ua.sportlife.customer dear bear"
    links = extract_links_from_text(text)
    link = links[0] if links[0:] else None
    if link:
        title = get_web_page_title(link)
        print(title)
        preview = get_preview(link)
        print(f"[{preview.title}]({link})\n![{preview.title}]({preview.image})\n\n")
